#include <err.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "wrapper.h"

static const size_t defaultsz = 10;

static const char *def_beg = "\tlet basis = Basis.nil in\n";
static const char *def_var = "\tlet (%s, basis) = Basis.add basis \"%s\" in\n";
static const char *def_sz = "\tStat.varNbUp %u;\n";

void env_new(struct env *e) {
	struct var_t *v;

	v = (struct var_t *)calloc(defaultsz, sizeof *v);
	if(v == NULL)
		err(1, "malloc(3)");

	e->max = defaultsz;
	e->used = 0;
	e->envs_sz = 0;
	e->next = 0;
	e->vars = v;
}

void env_del(struct env *e) {
	unsigned int i;

	for(i = 0; i < e->used; i++)
		free((void *)e->vars[i].src);

	e->max = 0;
	e->used = 0;
	e->envs_sz = 0;

	free(e->vars);
	e->vars = NULL;
}

void env_merge(struct env *e, ap_environment_t *ne) {
	unsigned int i;

	for(i = 0; i < e->envs_sz; i++)
		if(e->envs[i] == ne)
			return;

	if(e->envs_sz >= MAX_ENV)
		errx(1, "too many environments merged");

	e->envs[e->envs_sz] = ne;
	e->envs_sz++;
}

size_t varlen(struct env *e, unsigned int dim) {
	char b[50];
	unsigned int i;
	size_t rv;
	struct var_t *nv;
	ap_var_t av, av2;

	if(e->envs_sz == 0)
		errx(1, "varlen(): no environment");

	for(i = 0; i < e->envs_sz; i++)
		if(dim >= e->envs[i]->intdim)
			errx(1, "varlen(): real dimension");

	av = e->envs[0]->var_of_dim[dim];
	for(i = 1; i < e->envs_sz; i++) {
		av2 = e->envs[i]->var_of_dim[dim];
		if(ap_var_operations->compare(av, av2) != 0) {
			char *s, *s2;
			s = ap_var_operations->to_string(av);
			s2 = ap_var_operations->to_string(av2);
			errx(1, "varlen(): environments differ. \"%s\" <> \"%s\"", s, s2);
		}
	}

	for(i = 0; i < e->used; i++)
		if(e->vars[i].dim == dim)
			return strlen(e->vars[i].name);

	if(++e->used >= e->max) {
		struct var_t *nvars;
		size_t nmax = e->max * 2;

		nvars = (struct var_t *)realloc(e->vars, nmax * sizeof *nvars);
		if(nvars == NULL)
			err(1, "realloc(3)");

		e->vars = nvars;
		e->max = nmax;
	}

	rv = sprintf(b, "v%u", e->next++);
	if(rv >= 16)
		errx(1, "varlen(): too long name");

	nv = &e->vars[i];
	nv->dim = dim;
	nv->src = ap_var_operations->to_string(av);
	strcpy(nv->name, b);

	return rv;
}

char *varpretty(char *s, const struct env *e, unsigned int dim) {
	unsigned int i;

	for(i = 0; i < e->used; i++)
		if(e->vars[i].dim == dim)
			return frag_pretty(s, e->vars[i].name);

	errx(1, "varpretty()");
}

size_t
env_len(unsigned int flags, const struct env *e)
{
	char b[100];
	size_t l;
	unsigned int i;

	l = strlen(def_beg);
	for(i = 0; i < e->used; i++)
		l += sprintf(b, def_var, e->vars[i].name, e->vars[i].src);

	if(flags & L_SZ)
		l += sprintf(b, def_sz, e->used);

	return l;
}

char *
env_pretty(char *s, unsigned int flags, const struct env *e)
{
	unsigned int i;

	s = frag_pretty(s, def_beg);

	for(i = 0; i < e->used; i++)
		s += sprintf(s, def_var, e->vars[i].name, e->vars[i].src);

	if(flags & L_SZ)
		s += sprintf(s, def_sz, e->used);

	return s;
}
