#include <err.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#include "wrapper.h"

static const char *poldef = "\tlet p%u = Pol.mk basis [\n";
static const char *stats = "\tlet p%uEqs = %u and p%uIneqs = %u in\n";
static const char *szStr = "\tStat.consNbUp %i;\n";

static void to_lincons1(ap_lincons1_t *, ap_lincons0_t *, ap_environment_t *);

unsigned int polEqs(ap_lincons1_array_t *a) {
	unsigned int j, eqs = 0;
	ap_lincons0_t *l;

	l = a->lincons0_array.p;
	for(j = 0; j < a->lincons0_array.size; j++)
		if(l[j].linexpr0 != NULL || l[j].scalar != NULL)
			if(cons_typ(l[j].constyp) == L_EQ)
				eqs++;

	return eqs;
}

unsigned int polIneqs(ap_lincons1_array_t *a) {
	unsigned int j, ineqs = 0;
	ap_lincons0_t *l;

	l = a->lincons0_array.p;
	for(j = 0; j < a->lincons0_array.size; j++)
		if(l[j].linexpr0 != NULL || l[j].scalar != NULL)
			if(cons_typ(l[j].constyp) == L_INEQ)
				ineqs++;

	return ineqs;
}

size_t
pollenML(unsigned int flags, unsigned int i, ap_lincons1_array_t *a,
	struct env *e)
{
	char buf[100];
	unsigned int j, eqs = 0, ineqs = 0;
	size_t s;
	ap_lincons1_t l1;
	ap_lincons0_t *l0;

	if(i >= 1000)
		errx(1, "index too big");

	env_merge(e, a->env);

	s = sprintf(buf, poldef, i);

	l0 = a->lincons0_array.p;
	for(j = 0; j < a->lincons0_array.size; j++) {
		if(l0[j].linexpr0 != NULL || l0[j].scalar != NULL) {
			s += strlen(";\n");
			to_lincons1(&l1, &l0[j], a->env);
			s += lcons_len(&l1, e);

			switch(cons_typ(l0[j].constyp)) {
			case L_EQ: eqs++; break;
			case L_INEQ: ineqs++; break;
			default: errx(1, "pollenML()");
			}
		}
	}
	s += strlen("\n\t]\n\tin\n");

	if(flags & L_STAT)
		s += sprintf(buf, stats, i, eqs, i, ineqs);

	if(flags & L_SZ)
		s += sprintf(buf, szStr, eqs + ineqs);

	return s;
}

char *
polprettyML(char *s, unsigned int flags, unsigned int i,
	ap_lincons1_array_t *a, struct env *e)
{
	unsigned int j, eqs = 0, ineqs = 0;
	int need_sep = 0;
	ap_lincons1_t l1;
	ap_lincons0_t *l0;

	if(i >= 1000)
		errx(1, "index too big");

	s += sprintf(s, poldef, i);

	l0 = a->lincons0_array.p;
	for(j = 0; j < a->lincons0_array.size; j++) {
		if(l0[j].linexpr0 != NULL || l0[j].scalar != NULL) {
			if(need_sep)
				s = frag_pretty(s, ";\n");
			else
				need_sep = 1;

			to_lincons1(&l1, &l0[j], a->env);
			s = lcons_pretty(s, &l1, e);

			switch (cons_typ(l0[j].constyp)) {
			case L_EQ: eqs++; break;
			case L_INEQ: ineqs++; break;
			default: errx(1, "pollenML()");
			}
		}
	}
	s = frag_pretty(s, "\n\t]\n\tin\n");
	if(flags & L_STAT)
		s += sprintf(s, stats, i, eqs, i, ineqs);

	if(flags & L_SZ)
		s += sprintf(s, szStr, eqs + ineqs);

	return s;
}

static void
to_lincons1(ap_lincons1_t *l1, ap_lincons0_t *l0, ap_environment_t *e)
{
	l1->env = e;
	l1->lincons0 = *l0;
}
