#ifndef APRON_WRAPPER_H
#define APRON_WRAPPER_H

#include <gmp.h>
#include <stdio.h>
#include <time.h>

#include <sys/time.h>

#include "ap_global1.h"

#define MAX_ENV 2

struct var_t {
	unsigned int dim;
	const char *src;
	char name[16];
};

struct env {
	size_t max;
	size_t used;
	size_t envs_sz;
	unsigned int next;
	ap_environment_t *envs[MAX_ENV];
	struct var_t *vars;
};

struct record {
	unsigned int fcnt;
	unsigned int ffirst;
	unsigned int cnt;
	const char *callstr;
	FILE *f;
	const char *fprefix;
};

void env_new(struct env *);
void env_del(struct env *);
void env_merge(struct env *, ap_environment_t *);

size_t varlen(struct env *, unsigned int);
char *varpretty(char *, const struct env *, unsigned int);

size_t env_len(unsigned int, const struct env *);
char *env_pretty(char *, unsigned int, const struct env *);

#define C_NEG 1
#define C_PAREN 2
#define C_SKIP 4
#define L_STAT 8
#define L_SZ 16

#define L_EQ 0
#define L_INEQ 1

unsigned int cons_typ(ap_constyp_t);

size_t mpq_len(mpq_t, unsigned int);
char *mpq_pretty(char *, mpq_t, unsigned int);

size_t coeff_ml_len(unsigned int, const ap_coeff_t *);
char *coeff_ml_pretty(char *, unsigned int, const ap_coeff_t *);
void coeff2mpq(mpq_t, const ap_coeff_t *);

size_t lexpr_len(const ap_linexpr0_t *, struct env *);
char *lexpr_pretty(char *, const ap_linexpr0_t *, struct env *);

size_t lcons_len(const ap_lincons1_t *, struct env *);
char *lcons_pretty(char *, const ap_lincons1_t *, struct env *);

unsigned int tconsArEqs(ap_tcons0_array_t *);
unsigned int tconsArIneqs(ap_tcons0_array_t *);
size_t tcons_len(ap_environment_t *, ap_tcons0_t *, struct env *);
char *tcons_pretty(char *, ap_environment_t *, ap_tcons0_t *, struct env *);

size_t texpr_len(unsigned int, ap_environment_t *, ap_texpr0_t *,
	struct env *);
char *texpr_pretty(char *, unsigned int, ap_environment_t *, ap_texpr0_t *,
	struct env *);
void texpr_cst(mpq_t, const ap_texpr0_t *);

unsigned int polEqs(ap_lincons1_array_t *);
unsigned int polIneqs(ap_lincons1_array_t *);
size_t pollenML(unsigned int, unsigned int, ap_lincons1_array_t *,
	struct env *);
char *polprettyML(char *, unsigned int, unsigned int, ap_lincons1_array_t *,
	struct env *);

extern const char *inclcall;
void inclrec(struct record *, ap_manager_t *, double, bool,
	ap_abstract1_t *, ap_abstract1_t *);

extern const char *chullcall;
void chullrec(struct record *, ap_manager_t *, double,
	ap_abstract1_t *, size_t, ap_abstract1_t *);

extern const char *assigncall;
void assignrec(struct record *, ap_manager_t *, double,
	ap_abstract1_t *, ap_var_t *, ap_texpr1_t *, size_t, ap_abstract1_t *);

extern const char *meet_call;
void meetrec(struct record *, ap_manager_t *, double , ap_abstract1_t *,
	ap_tcons1_array_t *, ap_abstract1_t *);

/* Check that a manager points to the NewPolka abstract domain. */
bool check_manager(const ap_manager_t *);

/* Check that an abstract value points to the NewPolka abstract domain. */
bool check_abstract(const ap_abstract1_t *);

char *frag_pretty(char *, const char *);

double tstod(struct timespec *);
#endif
