type t = (Var.t * Cons.t) list

type rel_t =
| NoIncl
| Incl of (int * Cert.cstr_t) list

type meet_t =
| Ok of (int * t)
| Bottom of Cert.frag_t

let nil = []

let initc s = List.map (fun (x, c) -> (x, Cons.initc c)) s

let filter s c =
	let filter1 (x, c2) c1 =
		if Num.cmpz (Vec.get c1.Cons.c.Cstr.v x) = 0 then
			c1
		else
			Cons.elimc c1.Cons.id x c1 c2
 	in
	List.fold_right filter1 s c

let choose c =
	let rec find pred =
		function
		| Rtree.Nil -> None
		| Rtree.Sub (l, n, r) ->
			if pred n then
				Some (Var.XH, n)
			else
				match find pred l with
				| Some (x, n1) -> Some (Var.XO x, n1)
				| None ->
					match find pred r with
					| Some (x, n1) -> Some (Var.XI x, n1)
					| None -> None
	in
	match find (fun n -> Num.cmpz n <> 0) c.Cons.c.Cstr.v with
	| None -> failwith "EqSet.choose"
	| Some (x, a) -> (x, a)

let add s c =
	match c.Cons.c.Cstr.typ with
	| Cmp.Le | Cmp.Lt -> failwith "EqSet.add"
	| Cmp.Eq ->

	(* XXX: Is this check necessary? *)
	match Cstr.tellProp c.Cons.c with
	| Cstr.Trivial -> Ok (c.Cons.id, s)
	| Cstr.Contrad -> Bottom (Cert.mkF [c.Cons.id, Num.u])
	| Cstr.Nothing ->

	let c1 = filter s c in
	match Cstr.tellProp c1.Cons.c with
	| Cstr.Trivial -> Ok (c.Cons.id, s)
	| Cstr.Contrad -> Bottom c1.Cons.f
	| Cstr.Nothing ->
		let (x, ax) = choose c1 in
		let c2 = { c1 with
			Cons.c = Cstr.mult (Num.inv ax) c1.Cons.c;
			Cons.f = Cert.mapF (Num.mult (Num.inv ax)) c1.Cons.f }
		in
		Ok (c.Cons.id + 1, (x, c2)::s)

let addf s c = c::s

let implies s c =
	match Cstr.tellProp (filter s c).Cons.c with
	| Cstr.Trivial -> true
	| Cstr.Contrad (* should tell something useful *)
	| Cstr.Nothing -> false

let isIncl s1 s2 =
	if List.length s1 < List.length s2 then
		NoIncl
	else

	let s = initc s1 in
	let rec _isIncl cert =
	function
	| [] -> Incl cert
	| (_, c)::t ->
		(* XXX: Could this be done without the -1 dance? *)
		let c1 = filter s (Cons.mk (-1) c.Cons.c) in
		match Cstr.tellProp c1.Cons.c with
		| Cstr.Contrad | Cstr.Nothing -> NoIncl
		| Cstr.Trivial ->
			let (d, w) = List.partition (fun (id, _) -> id = -1) c1.Cons.f in
			match d with
			| [_, a] ->
				let norm (i, a1) = (i, Num.div a1 (Num.neg a)) in
				let f = Cert.mkF (List.map norm w) in
				_isIncl ((c.Cons.id, Cert.Direct f)::cert) t

			| _ -> failwith "EqSet.isIncl"
	in
	_isIncl [] s2

let isEq s1 s2 =
	if List.length s1 = List.length s2 then
		List.for_all2 (fun (x1 ,c1) (x2, c2) ->
			x1 = x2 && Cstr.isInclSyn c1.Cons.c c2.Cons.c) s1 s2
	else
		false

let rename s fromX toY =
	let rename1 (x, c) =
		((if x = fromX then toY else x), Cons.rename fromX toY c)
	in
	List.map rename1 s

let pick msk e =
	let rec find m v =
		match m, v with
		| Rtree.Nil, _ | _, Rtree.Nil -> None
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			if n1 = None || Num.cmpz n2 = 0 then
				match find l1 l2 with
				| Some _ as v -> v
				| None -> find r1 r2
			else
				n1
	in
	find msk e.Cons.c.Cstr.v

let rec subst x i e =
	function
	| [] -> (i, [])
	| (x1, e1)::l1 ->
		if x1 = x then
			failwith "EqSet.subst"
		else
			let e2 =
				if Num.cmpz (Vec.get e1.Cons.c.Cstr.v x) = 0 then
					{ e1 with Cons.id = i }
				else
					Cons.elimc i x e e1
			in
			let (i1, l2) = subst x (i + 1) e l1 in
			(i1, (x1, e2)::l2)

let rec tryDefs i msk =
	function
	| [] -> (None, i, [])
	| (x, e)::l ->
		if Rtree.get None msk x = None then
			let (def, i1, l1) = tryDefs (i + 1) msk l in
			(def, i1, (x, { e with Cons.id = i })::l1)
		else
			let (i1, l1) = subst x i e l in
			(Some (e, x), i1, l1)

let trySubstM i msk l =
	let (def, i1, l1) = tryDefs i msk l in
	if def = None then
		let rec _trysubstm i msk =
			function
			| [] -> (None, i, [])
			| (x, e)::l ->

			match pick msk e with
			| None ->
				let e1 = { e with Cons.id = i } in
				let (def, i1, l1) = _trysubstm (i + 1) msk l in
				(def, i1, (x, e1)::l1)

			| Some x ->
				let (i1, l1) = subst x i e l in
				(Some (e, x), i1, l1)
		in
		_trysubstm i msk l
	else
		(def, i1, l1)

let trySubst x i l =
	let msk = Rtree.set None Rtree.Nil x (Some x) in
	let (optx, i1, s1) = trySubstM i msk l in
	match optx with
	| None -> (None, i1, s1)
	| Some (e, _) -> (Some e, i1, s1)

let setupJoin dup b reloc alpha idOff s =
	let apply (x, c) (b1, reloc1, s1) =
		let (b2, reloc2, c1) = Cons.setupJoin dup b1 reloc1 alpha idOff c in
		let x1 =
			if dup then
				x
			else
				match Rtree.get None reloc2 x with
				| None -> failwith "EqSet.setupJoin"
				| Some x1 -> x1
		in
		(b2, reloc2, (x1, c1)::s1)
	in
(* List.fold_right is necessary because order needs to be preserved (echelon form) *)
	List.fold_right apply s (b, reloc, nil)

let extractc s = List.map (fun (_, c) -> Cons.extractc c) s

let list s = List.map (fun (_, c) -> c) s
let isTop s = (s = [])

let pr b e =
	List.fold_right (fun (_, c) s -> s ^ (Cons.pr b c) ^ "\n") e ""

let prExt b e =
	let pr1 x c = (Cons.prExt b c) ^ " (def: " ^ (Basis.pr b x) ^ ")\n" in
	List.fold_right (fun (x, c) s -> s ^ (pr1 x c)) e ""
