(** {! Pol} treats equalities and inequalities differently.
This module is the data structure that hosts the equalities of a polyhedron. *)

type t = (Var.t * Cons.t) list

type rel_t =
| NoIncl
| Incl of (int * Cert.cstr_t) list

type meet_t =
| Ok of (int * t)
| Bottom of Cert.frag_t

val nil: t

val filter: t -> Cons.t -> Cons.t

val choose: Cons.t -> Var.t * Num.t

val add: t -> Cons.t -> meet_t

(* fast check-less add *)
val addf: t -> (Var.t * Cons.t) -> t

val implies: t -> Cons.t -> bool
val isIncl: t -> t -> rel_t
val isEq: t -> t -> bool

val rename: t -> Var.t -> Var.t -> t

val pick: Var.t option Rtree.t -> Cons.t -> Var.t option
val subst: Var.t -> int -> Cons.t -> t -> int * t

val tryDefs: int -> Var.t option Rtree.t -> t ->
	(Cons.t * Var.t) option * int * t

val trySubstM: int -> Var.t option Rtree.t -> t ->
	(Cons.t * Var.t) option * int * t

val trySubst: Var.t -> int -> t -> Cons.t option * int * t

val setupJoin: bool -> Basis.t -> Var.t option Rtree.t -> Var.t -> int -> t ->
	Basis.t * Var.t option Rtree.t * t

val extractc: t -> (int * Cert.cstr_t) list
val initc: t -> t

val list: t -> Cons.t list
val isTop: t -> bool
val pr: Basis.t -> t -> string
val prExt: Basis.t -> t -> string
