let zToPos (z0: Num.Z.t): BinPos.positive =
	if Num.Z.cmp Num.Z.z z0 >= 0 then
		invalid_arg "Glue.zToPos"
	else
		let rec f (z: Num.Z.t): BinPos.positive =
			if Num.Z.cmp z Num.Z.u = 0 then
				BinPos.Coq_xH
			else
				if Num.Z.cmp (Num.Z.lAnd z Num.Z.u) Num.Z.u = 0 then
					BinPos.Coq_xI (f (Num.Z.shiftR z 1))
				else
					BinPos.Coq_xO (f (Num.Z.shiftR z 1))
		in
		f z0

let zToZ (z: Num.Z.t): BinInt.coq_Z =
	let sign = Num.Z.cmp Num.Z.z z in
	if sign = 0 then
		BinInt.Z0
	else
		if sign < 0 then
			BinInt.Zpos (zToPos z)
		else
			BinInt.Zneg (zToPos (Num.Z.neg z))

let nToNum (n: Num.t): NumC.t =
	let (num, den) = Num.toZ n in
	let q = {
		QArith_base.coq_Qnum = zToZ num;
		QArith_base.coq_Qden = zToPos den }
	in
	Qcanon.coq_Q2Qc q

let rec vToVec (v: Vec.t): VecC.t =
	match v with
	| Rtree.Nil -> VecC.Nil
	| Rtree.Sub (l, n, r) -> VecC.Sub (vToVec l, nToNum n, vToVec r)

let cToCons (c: Cstr.t): ConsC.t = {
	ConsC.typ = c.Cstr.typ;
	ConsC.coefs = vToVec c.Cstr.v;
	ConsC.cst = nToNum c.Cstr.c }

let pToPol (l: (int * Cstr.t) list): PolC.t =
	List.map (fun (i, c) -> (i, cToCons c)) l

let sort a =
	let cmp (i1, _) (i2, _) =
		match Pervasives.compare i1 i2 with
		| 0 -> invalid_arg "Glue.check"
		| n -> n
	in
	List.sort cmp a

let rec cToCert (c: Cert.t): PolC.cert_t =
	let convF (f: Cert.frag_t): CertC.frag_t =
		List.map (fun (i, n) -> (i, nToNum n)) f
	in
	let convC: Cert.cstr_t -> CertC.t = function
		| Cert.Direct f -> CertC.Direct (convF f)
		| Cert.SplitEq (f1, f2) -> CertC.SplitEq (convF f1, convF f2)
	in
	let convL (l: (int * Cert.cstr_t) list) =
		sort (List.map (fun (i, c) -> (i, convC c)) l)
	in
	match c with
	| Cert.Implies l -> PolC.Implies (convL l)
	| Cert.Empty f -> PolC.Empty (convF f)
	| Cert.Bind (i, c, ce1, ce2) -> PolC.Bind (i, cToCons c, convC ce1, cToCert ce2)


let check (l1: (int * Cstr.t) list) (l2: (int * Cstr.t) list) (c: Cert.t) =
	PolC.check (pToPol l1) (pToPol (sort l2)) (cToCert c)

let checkMeet (l1: (int * Cstr.t) list) (i, c) (l2: (int * Cstr.t) list option) (ce: Cert.t) =
	let p2 =
		match l2 with
		| None -> None
		| Some l -> Some (pToPol (sort l))
	in
	PolC.checkMeet (pToPol l1) (i, cToCons c) p2 (cToCert ce)
