(** {! Pol} treats equalities and inequalities differently.
This module is the data structure that hosts the inequalities of a polyhedron. *)

type t = Cons.t list

type prop_t =
| Empty of Cert.frag_t
| Trivial
| Implied of int * Cert.cstr_t
| Check of Cons.t

type rel_t =
| NoIncl
| Incl of (int * Cert.cstr_t) list

type meet_t =
| MeetOk of (int * t)
| NewEqs of (Cons.t * t * (Cons.t * Cert.cstr_t) list)
| MeetBot of Cert.frag_t

type simpl_t =
| SimplMin of t
| SimplBot of Cert.frag_t

val nil: t
val add: Basis.t -> EqSet.t -> t -> Cons.t -> meet_t

val isIncl: Basis.t -> EqSet.t -> t -> t -> rel_t
val isEq: t -> t -> bool
val rename: t -> Var.t -> Var.t -> t

val pick: Var.t option Rtree.t -> t -> Var.t option
val subst: Basis.t -> EqSet.t -> Var.t -> int -> Cons.t -> t -> int * t
val simpl: Basis.t -> EqSet.t -> t -> simpl_t
val fmElim: Basis.t -> EqSet.t -> Var.t -> int -> t -> int * t
val fmElimM: Basis.t -> EqSet.t -> int -> Var.t option Rtree.t -> t ->
	int * t

val setupJoin: bool -> Basis.t -> Var.t option Rtree.t -> Var.t -> int -> t ->
	Basis.t * Var.t option Rtree.t * t

(** [fromUnsat w i] turns an unsatisfiability witness [w] from the simplex algorithm
into a linear combination yielding the constraint identified by [i].
The simplex algorithm was given the complement (identified by [i]) of constraint [i].
The identifier [i] must appear exactly once in [w] and must have a non-zero coefficient;
exception [Invalid_argument "IneqSet.fromUnsat"] is raised otherwise. *)
val fromUnsat: Splx.witness_t -> int -> Cert.frag_t

val synIncl: EqSet.t -> t -> Cons.t -> prop_t

val extractc: t -> (int * Cert.cstr_t) list
val initc: t -> t

val list: t -> Cons.t list
val isTop: t -> bool
val pr: Basis.t -> t -> string
val prExt: Basis.t -> t -> string
