Require Import Bool.
Require NumC.
Require VecC.
Require Cmp.

Record t: Set := mk { coefs: VecC.t; typ: Cmp.t; cst: NumC.t }.

Definition Sat (c: t) (x: VecC.t): Prop :=
  (Cmp.denote (typ c)) (VecC.eval (coefs c) x) (cst c).

Definition Incl (c1: t) (c2: t): Prop :=
  forall x: VecC.t, Sat c1 x -> Sat c2 x.

Definition Contrad (c: t): Prop :=
  forall x: VecC.t, not (Sat c x).

Definition isInclEq (c1 c2: t) (Ht1: typ c1 = Cmp.Eq)
    (Hcoefs: VecC.Eq (coefs c1) (coefs c2)): Exc (Incl c1 c2).
  refine
    (match (typ c2) as t2 return typ c2 = t2 -> Exc (Incl c1 c2) with
     | Cmp.Eq =>
       match NumC.EqDec (cst c1) (cst c2) with
       | left _ => (fun Ht2 => value _)
       | right _ => (fun _ => error)
       end
     | Cmp.Lt =>
       match NumC.LtLeDec (cst c1) (cst c2) with
       | left _ => (fun Ht2 => value _)
       | right _ => (fun _ => error)
       end
     | Cmp.Le =>
       match NumC.LtLeDec (cst c2) (cst c1) with
       | left _ => (fun _ => error)
       | right _ => (fun Ht2 => value _)
       end
     end (refl_equal (typ c2)));
  unfold Incl, Sat; unfold VecC.Eq in Hcoefs;
  intro; rewrite Hcoefs, Ht1, Ht2; simpl;
  intro Hcmp; rewrite Hcmp; assumption.
Defined.

Definition isInclLt (c1 c2: t) (Ht1: typ c1 = Cmp.Lt)
    (Hcoefs: VecC.Eq (coefs c1) (coefs c2)): Exc (Incl c1 c2).
  refine
    (match (typ c2) as t2 return typ c2 = t2 -> Exc (Incl c1 c2) with
    | Cmp.Eq => (fun _ => error)
    | Cmp.Lt =>
      match NumC.LtLeDec (cst c2) (cst c1) with
      | left _ => (fun _ => error)
      | right _ => (fun Ht2 => value _)
      end
    | Cmp.Le =>
      match NumC.LtLeDec (cst c2) (cst c1) with
      | left _ => (fun _ => error)
      | right _ => (fun Ht2 => value _)
      end
    end (refl_equal (typ c2)));
  unfold Incl, Sat; unfold VecC.Eq in Hcoefs;
  intro; rewrite Hcoefs, Ht1, Ht2; simpl; intro Hcmp;
  [ exact (NumC.LtLe _ _ (NumC.LtLeTrans _ _ _ Hcmp l))
  | exact (NumC.LtLeTrans _ _ _ Hcmp l)].
Defined.

Definition isInclLe (c1 c2: t) (Ht1: typ c1 = Cmp.Le)
    (Hcoefs: VecC.Eq (coefs c1) (coefs c2)): Exc (Incl c1 c2).
  refine
    (match (typ c2) as t2 return typ c2 = t2 -> Exc (Incl c1 c2) with
    | Cmp.Eq => (fun _ => error)
    | Cmp.Lt =>
      match NumC.LtLeDec (cst c1) (cst c2) with
      | left _ => (fun Ht2 => value _)
      | right _ => (fun _ => error)
      end
    | Cmp.Le =>
      match NumC.LtLeDec (cst c2) (cst c1) with
      | left _ => (fun _ => error)
      | right _ => (fun Ht2 => value _)
      end
    end (eq_refl (typ c2)));
  unfold Incl, Sat; unfold VecC.Eq in Hcoefs;
  intro; rewrite Hcoefs, Ht1, Ht2; simpl; intro Hcmp;
  [ exact (NumC.LeTrans _ _ _ Hcmp l)
  | exact (NumC.LeLtTrans _ _ _ Hcmp l)].
Defined.

Definition isInclDep (c1 c2: t): Exc (Incl c1 c2) :=
  let h := VecC.isEq (coefs c1) (coefs c2) in
  match h as _h return h = _h -> Exc (Incl c1 c2) with
  | true =>
    let Heq Htrue := VecC.IsEqSatSpec (coefs c1) (coefs c2) Htrue in
    match typ c1 as t1 return typ c1 = t1 -> h = true -> Exc (Incl c1 c2) with
    | Cmp.Eq => (fun Ht1 h => isInclEq c1 c2 Ht1 (Heq h))
    | Cmp.Lt => (fun Ht1 h => isInclLt c1 c2 Ht1 (Heq h))
    | Cmp.Le => (fun Ht1 h => isInclLe c1 c2 Ht1 (Heq h))
    end (eq_refl (typ c1))
  | false => (fun _ => error)
  end (refl_equal h).

Definition isIncl (c1 c2: t): bool :=
  match isInclDep c1 c2 with
  | value _ => true
  | error => false
  end.

Lemma IsInclSatSpec: forall c1 c2: t, isIncl c1 c2 = true -> Incl c1 c2.
  intros c1 c2; unfold isIncl; destruct (isInclDep c1 c2); intro h;
  [ assumption | inversion h ].
Qed.

Definition isContradDep (c: t): Exc (Contrad c).
  refine (
  match VecC.isEqDep (coefs c) VecC.Nil with
  | value e =>
    match typ c as t0 return typ c = t0 -> Exc (Contrad c) with
    | Cmp.Eq =>
      match NumC.EqDec NumC.z (cst c) with
      | left _ => (fun _ => error)
      | right p => (fun t0 => value _)
      end
    | Cmp.Lt =>
      match NumC.LtLeDec NumC.z (cst c) with
      | left _ => (fun _ => error)
      | right p => (fun t0 => value _)
      end
    | Cmp.Le =>
      match NumC.LtLeDec (cst c) NumC.z with
      | left p => (fun t0 => value _)
      | right _ => (fun _ => error)
      end
    end (refl_equal (typ c))
  | error => error
  end);
  unfold Contrad, Sat; unfold VecC.Eq in e; intro x;
  rewrite t0, (e x), VecC.EvalNil; unfold Cmp.denote;
  [ exact p | exact (NumC.LtNotLe _ _ p) | exact (NumC.LeNotLt _ _ p) ].
Defined.

Definition isContrad (c: t): bool :=
  match isContradDep c with
  | value _ => true
  | error => false
  end.

Lemma IsContradSatSpec: forall c: t, isContrad c = true -> Contrad c.
  unfold isContrad; intro c; destruct (isContradDep c);
  intro h; [ assumption | inversion h ].
Qed.

Definition nil: t := mk VecC.Nil Cmp.Eq NumC.z.

Lemma NilSatSpec: forall x: VecC.t, Sat nil x.
  unfold nil, Sat, VecC.eval; simpl; intro; reflexivity.
Qed.

Definition add (c1 c2: t): t :=
  mk (VecC.add (coefs c1) (coefs c2)) (Cmp.add (typ c1) (typ c2))
     (NumC.add (cst c1) (cst c2)).

Lemma AddSatSpec: forall (c1 c2: t) (x: VecC.t), Sat c1 x -> Sat c2 x -> Sat (add c1 c2) x.
  intros c1 c2 x; unfold Sat, add; simpl;
  rewrite VecC.AddSatSpec;
  destruct (typ c1); destruct (typ c2); simpl;
  [ exact (NumC.AddEqEq _ _ _ _)
  | exact (NumC.AddEqLe _ _ _ _)
  | exact (NumC.AddEqLt _ _ _ _)
  | rewrite NumC.AddComm, (NumC.AddComm (cst c1) (cst c2));
    intros h1 h2; exact (NumC.AddEqLe _ _ _ _ h2 h1)
  | exact (NumC.AddLeLe _ _ _ _)
  | exact (NumC.AddLeLt _ _ _ _)
  | rewrite NumC.AddComm, (NumC.AddComm (cst c1) (cst c2));
    intros h1 h2; exact (NumC.AddEqLt _ _ _ _ h2 h1)
  | rewrite NumC.AddComm, (NumC.AddComm (cst c1) (cst c2));
    intros h1 h2; exact (NumC.AddLeLt _ _ _ _ h2 h1)
  | exact (NumC.AddLtLt _ _ _ _)].
Qed.

Lemma AddNilL: forall c: t, add nil c = c.
  unfold add; simpl; destruct 0.
  rewrite NumC.AddZL, VecC.AddZL.
  reflexivity.
Qed.

Definition mul (c: t) (n: { n0 | typ c = Cmp.Eq \/ not (NumC.Le n0 NumC.z) }): t :=
  mk (VecC.mul (proj1_sig n) (coefs c)) (typ c) (NumC.mul (proj1_sig n) (cst c)).

Lemma MulSatSpec: forall (c: t) (n: { n0 | typ c = Cmp.Eq \/ not (NumC.Le n0 NumC.z) })
    (x: VecC.t), Sat c x -> Sat (mul c n) x.
  intros c [n hn] x; unfold Sat, mul; simpl;
  destruct hn as [hn | hn];
  [ rewrite hn; simpl; rewrite (VecC.MulSatSpec n (coefs c) x); intro h; rewrite h; trivial
  | destruct (typ c); simpl; rewrite (VecC.MulSatSpec n (coefs c) x);
    [ intro h; rewrite h; trivial | exact (NumC.MulLe _ _ _ hn) | exact (NumC.MulLt _ _ _ hn)]].
Qed.

Definition splitEq (c: { c0: t | typ c0 = Cmp.Eq }): t * t :=
  let c1 := proj1_sig c in
  (mk (coefs c1) Cmp.Le (cst c1), mk (VecC.neg (coefs c1)) Cmp.Le (NumC.neg (cst c1))).

Lemma SplitEqSatSpec: forall (c: { c0: t | typ c0 = Cmp.Eq }) (c1 c2: t),
  fst (splitEq c) = c1 -> snd (splitEq c) = c2 -> forall x: VecC.t,
    Sat c1 x -> Sat c2 x -> Sat (proj1_sig c) x.
  unfold Sat; simpl.
  intros [c Hc] c1 c2 Hc1 Hc2 x.
  rewrite <- Hc1, <- Hc2; simpl.
  rewrite Hc; simpl.
  rewrite (VecC.NegSatSpec (coefs c) x).
  exact (NumC.SplitEq _ _).
Qed.
