let (x, b) = Basis.add Basis.nil "x"
let (y, b) = Basis.add b "y"

let mk l = List.map (fun (i, a) -> (i, Num.mk a)) l

let mkc t v c =
	Cstr.mk t (List.map (fun (c, v) -> (Num.mk c, v)) v) (Num.mk c)

let eq = mkc Cmp.Eq and le = mkc Cmp.Le

(* Cert.isEqF *)
let isEqF_ts =
	let chk (t, l1, l2, r) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let f1 = Cert.mkF (mk l1) and f2 = Cert.mkF (mk l2) in
		OUnit.assert_equal r (Cert.isEqF f1 f2)))
	in
	let tcs = [
		"empty", [], [], true;
		"one_empty0", [], [1, 1], false;
		"one_empty1", [1, 1], [], false;
		"one_same0", [1, 1], [1, 1], true;
		"one_same1", [2, 1], [2, 1], true;
		"one_same2", [1, 2], [1, 2], true;
		"one_other0", [1, 1], [2, 1], false;
		"one_other1", [1, 1], [1, 2], false;
		"many_same0", [1, 2; 2, 3], [1, 2; 2, 3], true;
		"many_order0", [2, 3; 1, 2], [1, 2; 2, 3], true;
		"many_other0", [1, 2; 2, 3], [1, 2; 3, 3], false;
		"many_other1", [1, 2; 2, 3], [3, 2; 2, 3], false;
		"many_other2", [1, 2; 2, 3], [1, 1; 2, 3], false;
		"many_other3", [1, 2; 2, 2], [1, 2; 2, 3], false;
	] in
	OUnit.TestList (List.map chk tcs)

(*	Cert.mkF *)
let mkF_override_ts =
	let chk (t, l) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let f = mk l in
		OUnit.assert_raises (Invalid_argument "Cert.cmp")
			(fun () -> Cert.mkF f)))
	in
	let tcs = [
		"override0", [0, 1; 0, 1];
		"override1", [0, 1; 0, 2];
		"override2", [0, 1; 1, 1; 1, 3];
		"override3", [0, 1; 1, 2; 0, 2]
	] in
	OUnit.TestList (List.map chk tcs)

let mkF_zero_ts =
	let chk (t, l) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let f = mk l in
		OUnit.assert_raises (Invalid_argument "Cert.mkF")
			(fun () -> Cert.mkF f)))
	in
	let tcs = [
		"z0", [0, 0];
		"z1", [0, 1; 1, 0];
		"z2", [1, 2; 0, 2; 3, 0]
	] in
	OUnit.TestList (List.map chk tcs)

let mkF_ts = OUnit.TestList [
		OUnit.TestLabel ("override", mkF_override_ts);
		OUnit.TestLabel ("zero", mkF_zero_ts)
	]

(* Cert.mergeF *)
let mergeF_ts =
	let chk (t, ll, r) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let lf = List.map (fun (l, i) -> (Cert.mkF (mk l), Num.mk i)) ll in
		let rf = Cert.mkF (mk r) and f = Cert.mergeF lf in
		let estr =
			"\nExpected: " ^ (Cert.prF rf) ^
			"\nGot: " ^ (Cert.prF f) ^
			"\nAs the result of the sum of the following lines:\n" ^
			(List.fold_left (fun s (f, i) -> s ^ "\t" ^
				(Num.pr i) ^ " * [" ^ (Cert.prF f) ^ "]\n") "" lf)
		in
		OUnit.assert_bool estr (Cert.isEqF rf f)))
	in
	let tcs = [
		"empty0", [], [];
		"empty1", [[], 0], [];
		"empty2", [[], 0; [], 0; [], 0], [];
		"empty3", [[], 1; [], 2; [], 3], [];
		"empty4", [[0, 1], 0; [], 1], [];
		"one0", [[0, 1], 1; [], 1], [0, 1];
		"one1", [[0, 1], 2; [], 1], [0, 2];
		"one2", [[0, 2], 2; [], 1], [0, 4];
		"one3", [[1, 2; 0, 1], 2; [], 1], [1, 4; 0, 2];
		"disjoined0", [[1,1; 2,1], 1; [3,1; 0,1], 1], [1,1; 2,1; 3,1; 0,1];
		"disjoined1", [[1,1; 2,1], 2; [3,1; 0,1], 1], [1,2; 2,2; 3,1; 0,1];
		"disjoined2", [[1,1; 2,1], 2; [3,1; 0,1], 3], [1,2; 2,2; 3,3; 0,3];
		"overlap0", [[1,1; 2,1], 2; [2,1; 0,1], 3], [1,2; 2,5; 0,3]
	] in
	OUnit.TestList (List.map chk tcs)

(* Cert.inlineF *)
let inlineF_ts =
	let mkF i = Cert.mkF (mk i) in
	let chk (t, fIn, fOut, defs) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_equal ~printer:Cert.prF ~cmp:Cert.isEqF fOut (Cert.inlineF defs fIn)))
	in
	let tcs = [
		"noSubst0", mkF [0, 1], mkF [0, 1], [];
		"noSubst1", mkF [0, 1], mkF [0, 1], [2, Cert.Direct (mkF [0, 1; 1, 2])];
		"noSubst2", mkF [0, 1; 1, 2], mkF [0, 1; 1, 2], [2, Cert.SplitEq (mkF [1, 2], mkF [1, -2])];
		"direct0", mkF [1, 1], mkF [0, 1], [1, Cert.Direct (mkF [0, 1])];
		"direct1", mkF [1, 1], mkF [0, 2], [1, Cert.Direct (mkF [0, 2])];
		"direct2", mkF [2, 1], mkF [0, 1; 1, 2], [2, Cert.Direct (mkF [0, 1; 1, 2])];
		"direct3", mkF [2, 2], mkF [0, 2; 1, 4], [2, Cert.Direct (mkF [0, 1; 1, 2])];
		"direct4", mkF [2, -2], mkF [0, -2; 1, -4], [2, Cert.Direct (mkF [0, 1; 1, 2])];
		"split0", mkF [2, 1], mkF [0, 1], [2, Cert.SplitEq (mkF [0, 1], mkF [1, 2])];
		"split1", mkF [2, -1], mkF [1, 2], [2, Cert.SplitEq (mkF [0, 1], mkF [1, 2])];
		"split2", mkF [2, 2], mkF [0, 2], [2, Cert.SplitEq (mkF [0, 1], mkF [1, 2])];
		"split3", mkF [2, -2], mkF [1, 4], [2, Cert.SplitEq (mkF [0, 1], mkF [1, 2])];
		"split4", mkF [2, 1], mkF [0, 1; 1, 2], [2, Cert.SplitEq (mkF [0, 1; 1, 2], mkF [1, 3])];
		"over0", mkF [0, 2; 1, 3], mkF [0, 8], [1, Cert.Direct (mkF [0, 2])];
		"both0", mkF [0, 2; 3, 1; 4, 3], mkF [0, 6], [
			3, Cert.SplitEq (mkF [0, 1], mkF [1, 2]);
			4, Cert.Direct (mkF [0, 1]) ];

		"norec0", mkF [0, 1; 2, 2], mkF [0, 1; 1, 4], [
			1, Cert.Direct (mkF [0, 1]);
			2, Cert.Direct (mkF [1, 2]) ]
	] in
	OUnit.TestList (List.map chk tcs)

(* Cert.inlineC *)
let inlineC_ts =
	let mkF i = Cert.mkF (mk i) in
	let d i = Cert.Direct (mkF i) and s i1 i2 = Cert.SplitEq (mkF i1, mkF i2) in
	let isEqC c1 c2 =
		match c1, c2 with
		| Cert.Direct _, Cert.SplitEq _ | Cert.SplitEq _, Cert.Direct _ -> false
		| Cert.Direct f1, Cert.Direct f2 -> Cert.isEqF f1 f2
		| Cert.SplitEq (f1, f2), Cert.SplitEq (f3, f4) -> Cert.isEqF f1 f3 && Cert.isEqF f2 f4
	in
	let chk (t, typ, cIn, cOut, defs) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_equal ~printer:Cert.prC ~cmp:isEqC cOut (Cert.inlineC defs typ cIn)))
	in
	let tcs = [
		"directLe0", Cmp.Le, d [0, 2; 2, 1], d [0, 3; 1, 2], [2, d [0, 1; 1, 2]];
		"directLt0", Cmp.Lt, d [0, 2; 2, 1], d [0, 3; 1, 2], [2, d [0, 1; 1, 2]];
		"directEq0", Cmp.Eq, d [2, -1; 1, -1], s [4, 1; 1, -1] [3, 1; 1, 1], [2, s [3, 1] [4, 1]];
		"splitEq0", Cmp.Eq, s [1, 1] [0, 2], s [3, 2; 4, 1] [2, 2], [1, d [3, 2; 4, 1]; 0, d [2, 1]]
	] in
	OUnit.TestList (List.map chk tcs)
		
let ts: OUnit.test = OUnit.TestList [
		OUnit.TestLabel ("isEqF", isEqF_ts);
		OUnit.TestLabel ("mkF", mkF_ts);
		OUnit.TestLabel ("mergeF", mergeF_ts);
		OUnit.TestLabel ("inlineF", inlineF_ts);
		OUnit.TestLabel ("inlineC", inlineC_ts)
	]
