let (x, b) = Basis.add Basis.nil "x"
let (y, b) = Basis.add b "y"
let (z, b) = Basis.add b "z"

let mkc t v c =
	Cstr.mk t (List.map (fun (c, v) -> (Num.mk c, v)) v) (Num.mk c)

let eq = mkc Cmp.Eq and le = mkc Cmp.Le and lt = mkc Cmp.Lt

let checknbasic s =
	let checkv st =
		(match st.Splx.low with
		| Some l -> Splx.Val.cmp l.Splx.bv st.Splx.v <= 0
		|None -> true) &&
		(match st.Splx.up with
		| Some u -> Splx.Val.cmp u.Splx.bv st.Splx.v >= 0
		|None -> true)
	in
	let rec _check m st =
		match m, st with
		| _, Rtree.Nil -> true
		| Rtree.Nil, Rtree.Sub (l, n, r) ->
			checkv n && _check Rtree.Nil l && _check Rtree.Nil r
		| Rtree.Sub (l1, Some _, r1), Rtree.Sub (l2, _, r2) ->
			_check l1 l2 && _check r1 r2
		| Rtree.Sub (l1, None, r1), Rtree.Sub (l2, n, r2) ->
			checkv n && _check l1 l2 && _check r1 r2
	in
		_check s.Splx.mat s.Splx.state

let checkmat s =
	let rec eval a ve st =
		match ve, st with
		| Rtree.Nil, _ | _, Rtree.Nil-> a
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			let incr = Splx.Val.add a (Splx.Val.muln n2.Splx.v n1) in
			eval (eval incr l1 l2) r1 r2
	in
	Rtree.fold
		(fun a n ->
			match n with
			| None -> a
			| Some ve ->
				if a then
					let v = eval Splx.Val.z ve s.Splx.state in
					Splx.Val.cmp Splx.Val.z v = 0
				else
					false)
		true s.Splx.mat

let checkinv s = checknbasic s && checkmat s

let ccheck s0 =
	let rec loop i s =
		let ns = Splx._check_body s in
		OUnit.assert_bool ((string_of_int i) ^ ": invariant") (checkinv ns);
		loop (i + 1) ns
	in
	let s = Splx.presimpl s0 in
	OUnit.assert_bool "presimpl: invariant" (checkinv s);
	try
		loop 1 s
	with
		Splx.Sat ns -> ns

let eqBnd ob1 ob2 =
	match ob1, ob2 with
	| None, None -> true
	| None, Some _ | Some _, None -> false
	| Some b1, Some b2 ->
		b1.Splx.id = b2.Splx.id &&
		Num.cmp b1.Splx.scale b2.Splx.scale = 0 &&
		Splx.Val.cmp b1.Splx.bv b2.Splx.bv = 0

let eqSt st1 st2 =
	Splx.Val.cmp st1.Splx.v st2.Splx.v = 0 &&
	eqBnd st1.Splx.low st2.Splx.low &&
	eqBnd st1.Splx.up st2.Splx.up

(* Splx.mergeB *)
let mergeB_okTs =
	let mkV i = Splx.Val.of_num (Num.mk i)
	and mkN i = Num.mk i
	and mkB id sc v = Some {Splx.id = id; Splx.scale = sc; Splx.bv = v} in
	let chk (t, st, stA, stB) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		try
			let actualSt = Splx.mergeB stA stB in
			OUnit.assert_equal ~cmp:eqSt ~printer:Splx.prSt st actualSt
		with
			Splx.Unsat _ -> OUnit.assert_failure "exception Splx.Unsat was raised."))
	in
	let tcs = [
		"nil0",
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None};

		"nil1", (* resulting value is that of the first argument *)
			{Splx.v = mkV 2; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 2; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None};

		"nil2", (* resulting value is that of the first argument *)
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 2; Splx.low = None; Splx.up = None};

		"noMergeLow0",
			{Splx.v = mkV 0; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = None},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = None};

		"noMergeLow1",
			{Splx.v = mkV 0; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = None},
			{Splx.v = mkV 0; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = None},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None};

		"noMergeUp0",
			{Splx.v = mkV 1; Splx.low = None; Splx.up = mkB 1 (mkN 1) (mkV 2)},
			{Splx.v = mkV 1; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = mkB 1 (mkN 1) (mkV 2)};

		"noMergeUp1",
			{Splx.v = mkV 0; Splx.low = None; Splx.up = mkB 1 (mkN 1) (mkV 2)},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = mkB 1 (mkN 1) (mkV 2)},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None};

		"noMerge0",
			{Splx.v = mkV 0; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = mkB 2 (mkN 1) (mkV 3)},
			{Splx.v = mkV 0; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = None},
			{Splx.v = mkV 1; Splx.low = None; Splx.up = mkB 2 (mkN 1) (mkV 3)};

		"noMerge1",
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = mkB 2 (mkN 1) (mkV 3)},
			{Splx.v = mkV 1; Splx.low = None; Splx.up = mkB 2 (mkN 1) (mkV 3)},
			{Splx.v = mkV 0; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = None};

		"mergeLow0", (* resulting value is out-of-bounds *)
			{Splx.v = mkV 2; Splx.low = mkB 2 (mkN 2) (mkV 3); Splx.up = None},
			{Splx.v = mkV 2; Splx.low = mkB 1 (mkN 1) (mkV 2); Splx.up = None},
			{Splx.v = mkV 3; Splx.low = mkB 2 (mkN 2) (mkV 3); Splx.up = None};

		"mergeLow1",
			{Splx.v = mkV 3; Splx.low = mkB 1 (mkN 1) (mkV 3); Splx.up = None},
			{Splx.v = mkV 3; Splx.low = mkB 1 (mkN 1) (mkV 3); Splx.up = None},
			{Splx.v = mkV 2; Splx.low = mkB 2 (mkN 2) (mkV 2); Splx.up = None};

		"mergeLow2", (* upper bound is preserved *)
			{Splx.v = mkV 3; Splx.low = mkB 1 (mkN 1) (mkV 3); Splx.up = mkB 3 (mkN 2) (mkV 4)},
			{Splx.v = mkV 3; Splx.low = mkB 1 (mkN 1) (mkV 3); Splx.up = mkB 3 (mkN 2) (mkV 4)},
			{Splx.v = mkV 2; Splx.low = mkB 2 (mkN 2) (mkV 2); Splx.up = None};

		"mergeLow3", (* upper bound is preserved *)
			{Splx.v = mkV 3; Splx.low = mkB 1 (mkN 1) (mkV 3); Splx.up = mkB 3 (mkN 1) (mkV 4)},
			{Splx.v = mkV 3; Splx.low = mkB 1 (mkN 1) (mkV 3); Splx.up = None},
			{Splx.v = mkV 2; Splx.low = mkB 2 (mkN 2) (mkV 2); Splx.up = mkB 3 (mkN 1) (mkV 4)};

		"mergeUp0",
			{Splx.v = mkV 2; Splx.low = None; Splx.up = mkB 2 (mkN 1) (mkV 2)},
			{Splx.v = mkV 2; Splx.low = None; Splx.up = mkB 1 (mkN 1) (mkV 3)},
			{Splx.v = mkV 1; Splx.low = None; Splx.up = mkB 2 (mkN 1) (mkV 2)};

		"mergeUp1",
			{Splx.v = mkV 2; Splx.low = None; Splx.up = mkB 1 (mkN 1) (mkV 2)},
			{Splx.v = mkV 2; Splx.low = None; Splx.up = mkB 1 (mkN 1) (mkV 2)},
			{Splx.v = mkV 1; Splx.low = None; Splx.up = mkB 2 (mkN 1) (mkV 3)};

		"mergeUp2", (* lower bound is preserved *)
			{Splx.v = mkV 2; Splx.low = mkB 3 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)},
			{Splx.v = mkV 2; Splx.low = mkB 3 (mkN 1) (mkV 1); Splx.up = mkB 1 (mkN 1) (mkV 3)},
			{Splx.v = mkV 1; Splx.low = None; Splx.up = mkB 2 (mkN 1) (mkV 2)};

		"mergeUp3", (* lower bound is preserved *)
			{Splx.v = mkV 2; Splx.low = mkB 3 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)},
			{Splx.v = mkV 2; Splx.low = None; Splx.up = mkB 1 (mkN 1) (mkV 3)},
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)};

		"merge0", (* both bounds from the same operand *)
			{Splx.v = mkV 2; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)},
			{Splx.v = mkV 2; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)},
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 0); Splx.up = mkB 4 (mkN 1) (mkV 4)};

		"merge1", (* both bounds from the same operand *)
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)},
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 0); Splx.up = mkB 4 (mkN 1) (mkV 4)},
			{Splx.v = mkV 2; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)};

		"merge2", (* one bound from each operand *)
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)},
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 1); Splx.up = mkB 4 (mkN 1) (mkV 4)},
			{Splx.v = mkV 2; Splx.low = mkB 1 (mkN 1) (mkV 0); Splx.up = mkB 2 (mkN 1) (mkV 2)};

		"merge3", (* one bound from each operand *)
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 4 (mkN 1) (mkV 2)},
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 0); Splx.up = mkB 4 (mkN 1) (mkV 2)},
			{Splx.v = mkV 2; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 4)};

		"merge4", (* keep previous if equal: lower bound *)
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 1); Splx.up = mkB 4 (mkN 1) (mkV 2)},
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 1); Splx.up = mkB 4 (mkN 1) (mkV 2)},
			{Splx.v = mkV 2; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 4)};

		"merge4", (* keep previous if equal: upper bound *)
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 4 (mkN 1) (mkV 2)},
			{Splx.v = mkV 1; Splx.low = mkB 3 (mkN 1) (mkV 0); Splx.up = mkB 4 (mkN 1) (mkV 2)},
			{Splx.v = mkV 2; Splx.low = mkB 1 (mkN 1) (mkV 1); Splx.up = mkB 2 (mkN 1) (mkV 2)}
	] in
	OUnit.TestList (List.map chk tcs)

let mergeB_koTs =
	let mkV i = Splx.Val.of_num (Num.mk i)
	and mkN i = Num.mk i
	and mkB id sc v = Some {Splx.id = id; Splx.scale = sc; Splx.bv = v} in
	let chk (t, w, stA, stB) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		try
			let _ = Splx.mergeB stA stB in
			OUnit.assert_failure "Splx.Unsat _ not raised"
		with
		| Splx.Unsat actualW ->
			let sort = List.sort (fun (i1, _) (i2, _) -> Pervasives.compare i1 i2)
			and eq = List.for_all2 (fun (i1, n1) (i2, n2) -> i1 = i2 && Num.cmp n1 n2 = 0) in
			try
				OUnit.assert_equal ~cmp:eq (sort w) (sort actualW)
			with
			| Invalid_argument _ -> OUnit.assert_failure "w and actualW don't have the same length"))
	in
	let tcs = [
		"already0", (* one of the two operands is self-contradictory *)
			[1, mkN 2; 2, mkN 1],
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 2) (mkV 2); Splx.up = mkB 2 (mkN 1) (mkV 1)},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None};

		"already1", (* one of the two operands is self-contradictory *)
			[1, mkN 2; 3, mkN 1],
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 2) (mkV 2); Splx.up = mkB 3 (mkN 1) (mkV 1)},
			{Splx.v = mkV 0; Splx.low = mkB 2 (mkN 1) (mkV 0); Splx.up = mkB 4 (mkN 1) (mkV 3)};

		"already2", (* one of the two operands is self-contradictory *)
			[1, mkN 2; 2, mkN 1],
			{Splx.v = mkV 0; Splx.low = None; Splx.up = None},
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 2) (mkV 2); Splx.up = mkB 2 (mkN 1) (mkV 1)};

		"already3", (* one of the two operands is self-contradictory *)
			[1, mkN 2; 3, mkN 1],
			{Splx.v = mkV 0; Splx.low = mkB 2 (mkN 1) (mkV 0); Splx.up = mkB 4 (mkN 1) (mkV 3)},
			{Splx.v = mkV 1; Splx.low = mkB 1 (mkN 2) (mkV 2); Splx.up = mkB 3 (mkN 1) (mkV 1)};

		"merge0", (* the merge leads to empty interval *)
			[1, mkN 3; 2, mkN 2],
			{Splx.v = mkV 0; Splx.low = mkB 1 (mkN 3) (mkV 2); Splx.up = None},
			{Splx.v = mkV 0; Splx.low = None; Splx.up = mkB 2 (mkN 2) (mkV 1)}
	] in
	OUnit.TestList (List.map chk tcs)

let mergeBTs = OUnit.TestList [
		OUnit.TestLabel ("ok", mergeB_okTs);
		OUnit.TestLabel ("ko", mergeB_koTs)
	]

(* Splx.setVal *)
let setValTs =
	let chk (t, res, ve, st) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let aRes = Splx.setVal ve st in
		let valEq v1 v2 = if Splx.Val.cmp v1 v2 = 0 then true else false in
		OUnit.assert_equal ~cmp:valEq ~printer:Splx.Val.pr res aRes))
	in
	let rtreeMk = Rtree.mk {Splx.v = Splx.Val.z; Splx.low = None; Splx.up = None}
	and stMk v = {Splx.v = v; Splx.low = None; Splx.up = None}
	and valMk i = Splx.Val.of_num (Num.mk i) in
	let tcs = [
		"nil0", valMk 0, Vec.nil, Rtree.Nil;
		"nil1", valMk 0, Vec.mk [Num.u, x], Rtree.Nil;
		"nil2", valMk 0, Vec.nil, rtreeMk [x, stMk (valMk 1)];
		"u0", valMk (-1), Vec.mk [Num.u, x], rtreeMk [x, stMk (valMk 1)];
		"u1", valMk (-2), Vec.mk [Num.u, x], rtreeMk [x, stMk (valMk 2)];
		"u2", valMk 1, Vec.mk [Num.mk (-1), x], rtreeMk [x, stMk (valMk 1)];
		"u3", valMk 2, Vec.mk [Num.u, x], rtreeMk [x, stMk (valMk (-2))];
		"m0", valMk 1, Vec.mk [Num.mk 2, x; Num.mk 3, y; Num.u, z],
			rtreeMk [x, stMk (valMk 1); y, stMk (valMk (-1))]
	] in
	OUnit.TestList (List.map chk tcs)

(* Splx.fitBnd *)
let fitBndTs =
	let mkV i = Splx.Val.of_num (Num.mk i)
	and mkB v = Some {Splx.id = 1; Splx.scale = Num.mk 0; Splx.bv = v} in
	let chkBnd (t, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (_, s1) = Splx.fitBnd s in
		OUnit.assert_equal ~cmp:eqBnd ~printer:Splx.prBnd s.Splx.low s1.Splx.low;
		OUnit.assert_equal ~cmp:eqBnd ~printer:Splx.prBnd s.Splx.low s1.Splx.low))
	and chkFit (t, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (_, s1) = Splx.fitBnd s in
		let satL =
			match s1.Splx.low with
			| None -> true
			| Some l -> Splx.Val.cmp l.Splx.bv s1.Splx.v <= 0
		and satU =
			match s1.Splx.up with
			| None -> true
			| Some u -> Splx.Val.cmp s1.Splx.v u.Splx.bv <= 0
		in
		OUnit.assert_bool (Splx.prSt s1) (satL && satU)))
	and chkDelta (t, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (d, s1) = Splx.fitBnd s in
		let vEq v1 v2 = (Splx.Val.cmp v1 v2 = 0) in
		OUnit.assert_equal ~cmp:vEq ~printer:Splx.Val.pr s1.Splx.v (Splx.Val.add s.Splx.v d)))
	in
	let tcs = [
		"none0", {Splx.v = mkV 1; Splx.low = None; Splx.up = None};
		"upSat0", {Splx.v = mkV 1; Splx.low = None; Splx.up = mkB (mkV 1)};
		"upSat1", {Splx.v = mkV 1; Splx.low = None; Splx.up = mkB (mkV 2)};
		"upFit0", {Splx.v = mkV 3; Splx.low = None; Splx.up = mkB (mkV 2)};
		"lowSat0", {Splx.v = mkV 1; Splx.low = mkB (mkV 1); Splx.up = None};
		"lowSat1", {Splx.v = mkV 1; Splx.low = mkB (mkV 0); Splx.up = None};
		"lowFit0", {Splx.v = mkV 0; Splx.low = mkB (mkV 1); Splx.up = None}
	] in
	OUnit.TestList [
		OUnit.TestLabel ("bnd", OUnit.TestList (List.map chkBnd tcs));
		OUnit.TestLabel ("fit", OUnit.TestList (List.map chkFit tcs));
		OUnit.TestLabel ("delta", OUnit.TestList (List.map chkDelta tcs))
	]

(* Splx.add *)
let addOkTs =
	let mkV i = Splx.Val.of_num (Num.mk i)
	and mkVn i = Splx.Val.ndelta (Num.mk i)
	and mkN i = Num.mk i
	and mkB id sc v = Some {Splx.id = id; Splx.scale = sc; Splx.bv = v} in
	let (sx0, b1) = Basis.add b "sx0" in
	let (_, b1) = Basis.add b1 "sx1" in
	let (_, b1) = Basis.add b1 "sx2" in
	let (sx3, b1) = Basis.add b1 "sx3" in
	let chkSt (t, i, c, s, _, v, l, u) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let s1 = Splx.add s (i, c) in
		let stV = Rtree.get {Splx.v = Splx.Val.z; Splx.low = None; Splx.up = None} s1.Splx.state v in
		OUnit.assert_equal ~cmp:eqBnd ~printer:Splx.prBnd l stV.Splx.low;
		OUnit.assert_equal ~cmp:eqBnd ~printer:Splx.prBnd u stV.Splx.up))
	and chkC (t, i, c, s, e, v, _, _) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let s1 = Splx.add s (i, c) in
		let e1 = Rtree.get None s1.Splx.mat v in
		match e with
		| None -> OUnit.assert_equal None e1
		| Some e2 ->
			match e1 with
			| None -> OUnit.assert_failure "constraint expected"
			| Some e3 -> OUnit.assert_equal ~cmp:Vec.isEq ~printer:(Vec.pr b1) e2 e3))
	and chkInv (t, i, c, s, _, _, _, _) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let s1 = Splx.add s (i, c) in
		let msg = Printf.sprintf "invariant\nbefore:\n%s\nafter:\n%s\n" (Splx.pr s) (Splx.pr s1) in
		OUnit.assert_bool msg (checkinv s1)))
	in
	let tcs = [
	(* no overriding *)
		(* single variable *)
		"sle0", 1, le [1, x] 1, Splx.mk b [],
			None, x, None, mkB 1 (mkN 1) (mkV 1);

		"sle1", 3, le [2, x] 4, Splx.mk b [],
			None, x, None, mkB 3 (Num.mk ~den:2 1) (mkV 2);

		"sle2", 4, le [2, x] (-4), Splx.mk b [],
			None, x, None, mkB 4 (Num.mk ~den:2 1) (mkV (-2));

		"sle3", 2, le [-2, x] (-4), Splx.mk b [],
			None, x, mkB 2 (Num.mk ~den:2 1) (mkV 2), None;

		"sle4", 1, le [1, x] 1, Splx.mk b [2, le [1, y] 1],
			None, x, None, mkB 1 (mkN 1) (mkV 1);

		"sle5", 1, le [1, x] 1, Splx.mk b [2, le [1, x; 1, y] 5],
			None, x, None, mkB 1 (mkN 1) (mkV 1);

		"sle6", 1, le [1, x] 1, Splx.mk b [2, le [-1, x] 1],
			None, x, mkB 2 (mkN 1) (mkV (-1)), mkB 1 (mkN 1) (mkV 1);

		"slt0", 2, lt [1, y] 1, Splx.mk b [],
			None, y, None, mkB 2 (mkN 1) (mkVn 1);

		"seq0", 2, eq [1, z] 1, Splx.mk b [],
			None, z, mkB 2 (mkN (-1)) (mkV 1), mkB 2 (mkN 1) (mkV 1);

		"sle7", 1, le [-1, x] (-1), Splx.mk b [
			2, eq [1, x; 1, z] 1;
			3, le [1, z; 1, y] 2;
			4, le [1, x; -1, y; 2, z] 2 ],
			None, x, mkB 1 (mkN 1) (mkV 1), None;

		(* multiple variables *)
		"mle0", 2, le [1, x; 1, y] 1, Splx.mk b [],
			Some (Vec.mk [mkN 1, sx0; mkN (-1), x; mkN (-1), y]),
			sx0, None, mkB 2 (mkN 1) (mkV 1);

		"mle1", 1, le [1, x; 1, y] 1, Splx.mk b [
			2, eq [1, x; 1, z] 1;
			3, le [1, z; 1, y] 2;
			4, le [1, x; -1, y; 2, z] 2 ],
			Some (Vec.mk [mkN 1, sx3; mkN (-1), x; mkN (-1), y]),
			sx3, None, mkB 1 (mkN 1) (mkV 1);

	(* overriding *)
		"overLe0", 1, le [1, x] 2, Splx.mk b [2, le [1, x] 3],
			None, x, None, mkB 1 (mkN 1) (mkV 2);

		"overEq0", 1, eq [2, y] 2, Splx.mk b [2, le [1, y] 2],
			None, y, mkB 1 (Num.mk ~den:2 (-1)) (mkV 1), mkB 1 (Num.mk ~den:2 1) (mkV 1);

	(* bug evidence *)
		"decompress0", 2, le [-1, x] (-17), Splx.mk b [
			0, le [-1, y] (-1);
			1, le [-1, x; 16, y] 15 ],
			None, x, mkB 2 (mkN 1) (mkV 17), None
	] in
	OUnit.TestList [
		OUnit.TestLabel ("state", OUnit.TestList (List.map chkSt tcs));
		OUnit.TestLabel ("cons", OUnit.TestList (List.map chkC tcs));
		OUnit.TestLabel ("inv", OUnit.TestList (List.map chkInv tcs))
	]

let addInvalTs =
	let chk (t, i, c, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_raises (Invalid_argument "Splx.add") (fun () -> Splx.add s (i, c))))
	in
	let tcs = [
		"triv0", 1, le [] 1, Splx.mk b [];
		"triv1", 1, lt [] 1, Splx.mk b [];
		"triv2", 1, eq [] 0, Splx.mk b [];
		"triv3", 1, le [] 1, Splx.mk b [2, le [1, x] 2];
		"contrad0", 1, le [] (-1), Splx.mk b [];
		"contrad1", 2, eq [] 1, Splx.mk b [1, le [1, x; 1, y] 2]
	] in
	OUnit.TestList (List.map chk tcs)

let addUnsatTs =
	let chk (t, i, c, s, w) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		try
			let _ = Splx.add s (i, c) in
			OUnit.assert_failure "Splx.Unsat _ not raised"
		with
		| Splx.Unsat actualW ->
			let sort = List.sort (fun (i1, _) (i2, _) -> Pervasives.compare i1 i2)
			and eq = List.for_all2 (fun (i1, n1) (i2, n2) -> i1 = i2 && Num.cmp n1 n2 = 0) in
			try
				OUnit.assert_equal ~cmp:eq (sort w) (sort actualW)
			with
			| Invalid_argument _ -> OUnit.assert_failure "w and actualW don't have the same length"))
	in
	let tcs = [
		"le0", 1, le [1, x] 1, Splx.mk b [2, le [-1, x] (-2)], [1, Num.u; 2, Num.u]
	] in
	OUnit.TestList (List.map chk tcs)

let addTs = OUnit.TestList [
		OUnit.TestLabel ("ok", addOkTs);
		OUnit.TestLabel ("inval", addInvalTs);
		OUnit.TestLabel ("unsat", addUnsatTs)
	]

(*	Splx.pickbasic *)
let _pick_basic_finished_ts: OUnit.test =
	let check ((title: string), (i: Splx.t)): OUnit.test =
		OUnit.TestLabel (title,
			OUnit.TestCase (function () ->
				OUnit.assert_equal (Splx.pickbasic i) None
			)
		) in
	let testcases: (string * Splx.t) list = [
		"empty", Splx.mk b [];
		"sat_bounds", Splx.mk b [0, eq [1, x] 0; 1, le [1, x; 1, y] 1]
	] in
	OUnit.TestList (List.map check testcases)

let _pick_basic_ts: OUnit.test = OUnit.TestList [
		OUnit.TestLabel ("finished", _pick_basic_finished_ts)
	]

(* Splx.pivot *)
let pivotTs =
	let (sx1, _) = Basis.add b "sx1" in
	let chkB (t, xB, xN, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let m = Splx.pivot s.Splx.mat xB xN in
		OUnit.assert_equal None (Rtree.get None m xB)))
	in
	let chkN (t, xB, xN, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let m = Splx.pivot s.Splx.mat xB xN in
		let nodeB = Rtree.get None s.Splx.mat xB and nodeN = Rtree.get None m xN in
		match nodeB, nodeN with
		| None, _ | _, None -> OUnit.assert_failure "not basic"
		| Some consB, Some consN ->
			match Vec.isomorph consB consN with
			| None -> OUnit.assert_failure "bad normalization"
			| Some n ->
				let numEq n1 n2 = (Num.cmp n1 n2 = 0) in
				OUnit.assert_equal ~cmp:numEq ~printer:Num.pr (Vec.get consB xN) n))
	in
	let chkProj (t, xB, xN, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let m = Splx.pivot s.Splx.mat xB xN in
		let cnt i = function
			| None -> i
			| Some cons ->
				if Num.cmpz (Vec.get cons xN) = 0 then
					i
				else
					i + 1
		in
		OUnit.assert_equal ~printer:string_of_int 1 (Rtree.fold cnt 0 m)))
	in
	let tcs = [
	"basic", sx1, x, Splx.mk b [
		0, le [-1, x; -1, y] (-1);
		1, le [1, x] 1;
		2, eq [-1, x; -1, z] 0;
		3, le [1, z] 2;
		4, le [1, y] 2];

	"coef0", sx1, x, Splx.mk b [
		0, le [-1, x; -1, y] (-1);
		1, le [1, x] 1;
		2, eq [-2, x; -1, z] 0;
		3, le [1, z] 2;
		4, le [1, y] 2];

	"coef1", sx1, x, Splx.mk b [
		0, le [-2, x; -1, y] (-1);
		1, le [1, x] 1;
		2, eq [-1, x; -1, z] 0;
		3, le [1, z] 2;
		4, le [1, y] 2];

	"noX0", sx1, x, Splx.mk b [
		0, le [-2, x; 1, y] (-1);
		1, le [1, y; 2, z] (-2);
		2, le [1, x] 1 ];

	"noX1", sx1, x, Splx.mk b [
		0, le [1, x; 1, z] (-1);
		1, le [1, y; 2, z] 3;
		2, le [1, x; 2, y] 3;
		3, le [1, x] 2 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("basic", OUnit.TestList (List.map chkB tcs));
		OUnit.TestLabel ("non-basic", OUnit.TestList (List.map chkN tcs));
		OUnit.TestLabel ("proj", OUnit.TestList (List.map chkProj tcs))
	]

let _check_body_unsat_ts: OUnit.test =
	let check ((title: string), (i: Splx.t)): OUnit.test =
		let err: string = "Splx.Unsat not raised" in
		OUnit.TestLabel (title,
			OUnit.TestCase (function () ->
				let res: bool = try let _ = Splx._check_body i in false
					with Splx.Unsat _ -> true in
				OUnit.assert_bool err res
			)
		) in
	let testcases: (string * Splx.t) list = [
	"unit0", Splx.mk b [
		0, le [1, x] 0;
		1, le [1, y] 0;
		2, eq [1, x; 1, y] 1]

	] in
	OUnit.TestList (List.map check testcases)

let _check_body_ts: OUnit.test = OUnit.TestList [
		OUnit.TestLabel ("unsat", _check_body_unsat_ts);
	]

let check_sat_ts =
	let check (t, s0) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_bool "unsat"
			(try
				let _ = Splx.check s0 in
				let _ = ccheck s0 in
				true
			with Splx.Unsat _ -> false)))
	in
	let tcs = [
		"decompress0", Splx.mk b [
			0, le [-1, y] (-1);
			1, le [-1, x; 16, y] 15;
			2, le [-1, x] (-17)];

		"decompress1", Splx.mk b [
			0, le [-1, y] (-1);
			1, le [-1, x; 16, y] (-18);
			2, le [-1, x] (-17)];

		"decompress2", Splx.mk b [
			0, le [-1, y] (-1);
			1, le [-1, x; 16, y] 15;
			2, le [-1, x] (-17);
			3, le [-1, x; 16, y] (-18)]
	] in
		OUnit.TestList (List.map check tcs)

let check_unsat_ts: OUnit.test =
	let chk (t, w, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let print_witness (l: Splx.witness_t): string =
			let pretty1 ((i: int), (s: Num.t)): string =
				(Num.pr s) ^ " * eq" ^ (string_of_int i) in
			let rec p (s: string) (l': Splx.witness_t): string =
				match l' with [] -> s
				| h::t -> if String.length s = 0 then p (pretty1 h) t
					else p (s ^ ", " ^ (pretty1 h)) t in
			"got " ^ (p "" l)
		in
		let check_witness (w: Splx.witness_t) (w': Splx.witness_t): bool =
			let p (acc: bool) ((i: int), (s: Num.t)): bool =
				let p' ((i': int), (s': Num.t)): bool =
					(i = i') && Num.cmp s s' = 0
				in
				acc && (try let _ = List.find p' w' in true
					with Not_found -> false)
			in
			List.fold_left p true w
		in
		let (err, res) =
			try
				let _ = Splx.check s in
				("sat", false)
			with
				Splx.Unsat f ->
					if check_witness w f then
						try
							let _ = ccheck s in
							("ccheck sat", false)
						with
							Splx.Unsat f ->
								if check_witness w f then
									("", true)
								else
									("ccheck unsat", false)
					else
						(print_witness f, false)
		in
		OUnit.assert_bool err res))
	in
	let tcs = [
		"simple_le0", [0, Num.u; 1, Num.u; 2, Num.mk 2], Splx.mk b [
			0, le  [-1, x] (-3);
			1, le [1, x; 2, y] 3;
			2, le [-1, y] (-1)];

		"simple_le1", [0, Num.u; 1, Num.u; 2, Num.u], Splx.mk b [
			0, le [-1, x] (-3);
			1, le [1, x; 2, y] 3;
			2, le [-2, y] (-2)];

		"simple_eq0", [0, Num.u; 1, Num.u; 2, Num.mk 2], Splx.mk b [
			0, le  [-1, x] (-3);
			1, eq [1, x; 2, y] 3;
			2, le [-1, y] (-1)];

		"simple_eq1", [0, Num.u; 1, Num.negU; 2, Num.mk 2], Splx.mk b [
			0, le [-1, x] (-3);
			1, eq [-1, x; -2, y] (-3);
			2, le [-1, y] (-1)];

		"simple_eq2", [0, Num.u; 1, Num.u; 2, Num.mk 2], Splx.mk b [
			0, le [-1, x] (-3);
			1, eq [1, x; 2, y] 3;
			2, le [-1, y] (-1)];

		"useless_l0", [0, Num.u; 2, Num.u; 3, Num.mk 2], Splx.mk b [
			0, le [-1, x] (-3);
			1, le [1, x; 1, y] 4;
			2, eq [1, x; 2, y] 3;
			3, le [-1, y] (-1)];

		"useless_l1", [0, Num.u; 2, Num.u; 3, Num.mk 2], Splx.mk b [
			0, le [-1, x] (-3);
			1, le [1, x] 4;
			2, eq [1, x; 2, y] 3;
			3, le [-1, y] (-1)]
	] in
	OUnit.TestList (List.map chk tcs)

let check_ts = OUnit.TestList
	[OUnit.TestLabel ("sat", check_sat_ts);
	OUnit.TestLabel ("unsat", check_unsat_ts)]

let compl_ok_ts =
	let getstate s x =
		let z = { Splx.v = Splx.Val.z; Splx.low = None; Splx.up = None } in
			Rtree.get z s.Splx.state x
	in
	let chk_inv (t, id, _, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let s = Splx.check s in
		if checkinv s then
			let s1 = Splx.compl s id in
			let e = "invariant\n" ^ (Splx.pr s1) in
			OUnit.assert_bool e (checkinv s1)
		else
			failwith "Splx_t.compl_ok_ts"))
	in
	let chk_same (t, id, x, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let s = Splx.check s in
		let s1 = Splx.compl s id in
		let xb0 = getstate s x and xb1 = getstate s1 x in
		match xb0.Splx.low with
		| Some l when l.Splx.id = id -> OUnit.assert_equal None xb1.Splx.low
		| _ ->
			match xb0.Splx.up with
			| Some u when u.Splx.id = id -> OUnit.assert_equal None xb1.Splx.up
			| _ -> failwith "Splx_t.compl_ok_ts"))
	in
	let chk_compl (t, id, x, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let s = Splx.check s in
		let s1 = Splx.compl s id in
		let xb0 = getstate s x and xb1 = getstate s1 x in
		let cmp v1 optv2 =
			match optv2 with
			| None -> ("None", false)
			| Some v2 ->
				let e = "Some (" ^ (Splx.Val.pr v2.Splx.bv) ^ ")" in
				(e, Splx.Val.cmp v1 v2.Splx.bv = 0)
		in
		match xb0.Splx.low with
		| Some l when l.Splx.id = id ->
			let (e, r) = cmp (Splx.Val.subdelta l.Splx.bv) xb1.Splx.up in
			OUnit.assert_bool e r

		| _ ->
			match xb0.Splx.up with
			| Some u when u.Splx.id = id ->
				let (e, r) = cmp (Splx.Val.adddelta u.Splx.bv) xb1.Splx.low in
				OUnit.assert_bool e r

			| _ -> failwith "Splx_t.compl_ok_ts"))
	in
	let (sx1, b1) = Basis.add b "sx1" in
	let (sx2, _) = Basis.add b1 "sx2" in
	let tcs = [
		"one0", 0, x, Splx.mk b [
			0, le [1, x] 1
		];

		"one0", 0, sx1, Splx.mk b [
			0, le [1, x; 2, y] 2
		];

		"m0", 1, y, Splx.mk b [
			0, le [1, x; 2, y] 2;
			1, le [1, y] 1;
			2, le [-1, x] 1
		];

		"m1", 2, sx2, Splx.mk b [
			0, le [1, x; 2, y] 2;
			1, le [1, y] 1;
			2, le [1, y; 2, z] 2;
			3, le [1, z] 30;
			4, le [-1, x] 1
		]

	] in
	OUnit.TestList [
		OUnit.TestLabel ("inv", OUnit.TestList (List.map chk_inv tcs));
		OUnit.TestLabel ("same", OUnit.TestList (List.map chk_same tcs));
		OUnit.TestLabel ("compl", OUnit.TestList (List.map chk_compl tcs))
	]

let compl_ko_ts =
	let chk (t, id, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let s = Splx.check s in
		OUnit.assert_raises (Invalid_argument "Splx.compl")
			(fun () -> Splx.compl s id)))
	in
	let tcs = [
		"empty", 0, Splx.mk b [];
		"one0", 1, Splx.mk b [
			0, le [1, x] 1];

		"one1", 1, Splx.mk b [
			0, le [1, x; 1, y] 1];

		"0", 2, Splx.mk b [
			0, le [1, x] 1;
			1, le [1, x; 1, y] 1];

		"hole0", 1, Splx.mk b [
			0, le [1, x] 1;
			2, le [1, x; 1, y] 1]
	] in
	OUnit.TestList (List.map chk tcs)

let compl_ts = OUnit.TestList [
		OUnit.TestLabel ("ko", compl_ko_ts);
		OUnit.TestLabel ("ok", compl_ok_ts)
	]

let stricten_ok_ts =
	let chk_inv (t, id, _, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let s1 = Splx.check s in
		if checkinv s1 then
			let s2 = Splx.stricten s1 id in
			let e = "invariant\n" ^ (Splx.pr s2) in
			OUnit.assert_bool e (checkinv s2)
		else
			failwith "Splx_t.stricten_ok_ts"))
	in
	let chk_bnd (t, id, x, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->

		let getstate s x =
			let z = { Splx.v = Splx.Val.z; Splx.low = None; Splx.up = None }
			in
				Rtree.get z s.Splx.state x
		in
		let s2 = Splx.stricten s id in
		let st1 = getstate s x and st2 = getstate s2 x in
		let (a, e) =
			let pr v1 v2 =
				"expected: " ^ (Splx.Val.pr v1) ^ " got: " ^ (Splx.Val.pr v2)
			in
			match st1.Splx.low, st2.Splx.low with
			| Some l1, Some l2 when l1.Splx.id = id ->
				if Splx.Val.hasDelta l1.Splx.bv then
					(false, "original bound has a delta")
				else
					let v = Splx.Val.adddelta l1.Splx.bv in
					(Splx.Val.cmp v l2.Splx.bv = 0, pr v l2.Splx.bv)

			| None, Some _ -> (false, "bound appeared")
			| Some _, None -> (false, "bound disappeared")
			| Some _, Some _ | None, None ->

			match st1.Splx.up, st2.Splx.up with
			| Some u1, Some u2 when u1.Splx.id = id ->
				if Splx.Val.hasDelta u1.Splx.bv then
					(false, "original bound has a delta")
				else
					let v = Splx.Val.subdelta u1.Splx.bv in
					(Splx.Val.cmp v u2.Splx.bv = 0, pr v u2.Splx.bv)

			| Some _, Some _ -> (false, "invalid id")
			| None, Some _ -> (false, "bound appeared")
			| Some _, None -> (false, "bound disappeared")
			| None, None -> (false, "no bound")
		in
		OUnit.assert_bool e a))
	in
	let (sx1, _) = Basis.add b "sx1" in
	let tcs = [
		"up0", 0, x, Splx.mk b [
			0, le [1, x] 0 ];

		"up1", 0, x, Splx.mk b [
			0, le [1, x] 1 ];

		"up2", 0, sx1, Splx.mk b [
			0, le [1, x; 1, y] 1 ];

		"up3", 0, x, Splx.mk b [
			0, le [1, x] (-1) ];

		"low0", 0, x, Splx.mk b [
			0, le [-1, x] 0 ];

		"m0", 1, sx1, Splx.mk b [
			0, le [1, x] 0;
			1, le [1, x; 1, y] 1;
			2, le [1, y; 1, z] 0;
			3, le [1, y] 15;
			4, le [1, z] 15 ];

		"m1", 0, x, Splx.mk b [
			0, le [1, x] 0;
			1, le [1, x; 1, y] 1;
			2, le [1, y; 1, z] 0 ]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("inv", OUnit.TestList (List.map chk_inv tcs));
		OUnit.TestLabel ("bnd", OUnit.TestList (List.map chk_bnd tcs))
	]

let stricten_inval_ts =
	let chk (t, id, s) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_raises (Invalid_argument "Splx.stricten")
			(fun () -> Splx.stricten s id)))
	in
	let tcs = [
	(* missing identifier *)
		"empty", 0, Splx.mk b [];
		"one0", 1, Splx.mk b [
			0, le [1, x] 1];

		"one1", 1, Splx.mk b [
			0, le [1, x; 1, y] 1];

		"0", 2, Splx.mk b [
			0, le [1, x] 1;
			1, le [1, x; 1, y] 1];

		"hole0", 1, Splx.mk b [
			0, le [1, x] 1;
			2, le [1, x; 1, y] 1];

	(* already strict bound *)
		"up0", 0, Splx.mk b [
			0, lt [1, x] 0 ];

		"up1", 0, Splx.mk b [
			0, lt [1, x] 1 ];

		"up2", 0, Splx.mk b [
			0, lt [1, x; 1, y] 1 ];

		"up3", 0, Splx.mk b [
			0, lt [1, x] (-1) ];

		"low0", 0, Splx.mk b [
			0, lt [-1, x] 0 ];

		"m0", 1, Splx.mk b [
			0, le [1, x] 0;
			1, lt [1, x; 1, y] 1;
			2, le [1, y; 1, z] 0;
			3, le [1, y] 15;
			4, le [1, z] 15 ];

		"m1", 0, Splx.mk b [
			0, lt [1, x] 0;
			1, le [1, x; 1, y] 1;
			2, le [1, y; 1, z] 0 ]
	] in
	OUnit.TestList (List.map chk tcs)

let stricten_unsat_ts =
	let chk (t, id, w, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let (a, e) =
		try
			let _ = Splx.stricten s id in
			(false, "Splx.Unsat not raised")
		with
			Splx.Unsat w1 -> (Cert.isEqF w w1, "bad cert" ^ (Cert.prF w1))
		in
		OUnit.assert_bool e a))
	in
	let tcs = [
		"simpl0", 0, [0, Num.u; 1, Num.u], Splx.mk b [
			0, le [1, x] 0;
			1, le [-1, x] 0 ];

		"scale0", 0, [0, Num.mk ~den:2 1; 1, Num.u], Splx.mk b [
			0, le [2, y] 0;
			1, le [-1, y] 0 ]
	] in
	OUnit.TestList (List.map chk tcs)

let stricten_ts =
	OUnit.TestList [
		OUnit.TestLabel ("ok", stricten_ok_ts);
		OUnit.TestLabel ("inval", stricten_inval_ts);
		OUnit.TestLabel ("unsat", stricten_unsat_ts)
	]

let forget_ok_ts =
	let chk_inv (t, id, _, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		if checkinv s then
			let s2 = Splx.forget s id in
			let e = "invariant\n" ^ (Splx.pr s2) in
			OUnit.assert_bool e (checkinv s2)
		else
			failwith "Splx_t.forget_ok_ts"))
	in
	let chk_bnd (t, id, x, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let getbnd s x = Rtree.get
			{ Splx.v = Splx.Val.z; Splx.low = None; Splx.up = None }
				s.Splx.state x
		in
		let chkbnd b1 b2 =
			match b1, b2 with
			| Some b, None when b.Splx.id = id -> true
			| Some b, Some _ when b.Splx.id = id -> false
			| None, None -> true
			| None, Some _ -> false
			| Some _, _ -> true
		in
		let s2 = Splx.forget s id in
		let b1 = getbnd s x and b2 = getbnd s2 x in
		let rup = chkbnd b1.Splx.up b2.Splx.up in
		let rlow = chkbnd b1.Splx.low b2.Splx.low in
		let e = "bound not forgotten\n" ^ (Splx.pr s2) in
		OUnit.assert_bool e (rlow && rup)))
	in
	let chk_mat (t, id, x, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let xbnd = Rtree.get
			{ Splx.v = Splx.Val.z; Splx.low = None; Splx.up = None }
				s.Splx.state x
		in
		if xbnd.Splx.up = None || xbnd.Splx.low = None then
			let count = Rtree.fold
				(fun n -> function None -> n | Some _ -> n + 1) 0 in
			let s2 = Splx.forget s id in
			let cnt1 = count s.Splx.mat and cnt2 = count s2.Splx.mat in
			let e = "cnt1 - cnt2 = " ^ (string_of_int (cnt1 - cnt2)) in
			OUnit.assert_bool e (cnt1 - cnt2 = 1)
		else
			()))
	in
	let tcs = [
		"one0", 0, x, Splx.mk b [
			0, le [1, x] 1;
			1, le [-1, x] 0];

		"cons0", 0, x, Splx.mk b [
			0, le [1, x] 1;
			1, le [1, x; 1, y] 0;
			2, le [1, x; 1, z] 0;
			3, le [1, y] 0;
			4, le [1, z] 0];

		"cons1", 0, x, Splx.mk b [
			0, le [1, x] 1;
			1, le [1, x; 1, y] 0]
	] in
	OUnit.TestList [
		OUnit.TestLabel ("inv", OUnit.TestList (List.map chk_inv tcs));
		OUnit.TestLabel ("bnd", OUnit.TestList (List.map chk_bnd tcs));
		OUnit.TestLabel ("mat", OUnit.TestList (List.map chk_mat tcs))
	]

let forget_ko_ts =
	let chk (t, id, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_raises (Invalid_argument "Splx.forget")
			(fun () -> Splx.forget (Splx.check s) id)))
	in
	let tcs = [
		"empty", 0, Splx.mk b [];
		"one0", 1, Splx.mk b [
			0, le [1, x] 1];

		"one1", 1, Splx.mk b [
			0, le [1, x; 1, y] 1];

		"0", 2, Splx.mk b [
			0, le [1, x] 1;
			1, le [1, x; 1, y] 1];

		"hole0", 1, Splx.mk b [
			0, le [1, x] 1;
			2, le [1, x; 1, y] 1]
	] in
	OUnit.TestList (List.map chk tcs)

let forget_ts = OUnit.TestList [
		OUnit.TestLabel ("ok", forget_ok_ts);
		OUnit.TestLabel ("ko", forget_ko_ts)
	]

let ts: OUnit.test = OUnit.TestList [
		OUnit.TestLabel ("mergeB", mergeBTs);
		OUnit.TestLabel ("setVal", setValTs);
		OUnit.TestLabel ("fitBnd", fitBndTs);
		OUnit.TestLabel ("add", addTs);
		OUnit.TestLabel ("_pick_basic", _pick_basic_ts);
		OUnit.TestLabel ("pivot", pivotTs);
		OUnit.TestLabel ("_check_body", _check_body_ts);
		OUnit.TestLabel ("check", check_ts);
		OUnit.TestLabel ("compl", compl_ts);
		OUnit.TestLabel ("stricten", stricten_ts);
		OUnit.TestLabel ("forget", forget_ts)
	]
