(* test suite for Splx.Val *)

(* Splx.Val.z *)
let z_ts =
	OUnit.TestCase (fun () ->
		let z1 = Splx.Val.z and z2 = Splx.Val.of_num Num.z in
		let e = (Splx.Val.pr z1) ^ " <> " ^ (Splx.Val.pr z2) in
		OUnit.assert_bool e (Splx.Val.cmp z1 z2 = 0))

(* Splx.Val.cmp *)
let cmp_ts =
	let eq = 0 and lt = -1 and gt = 1 in
	let chk (t, v1, v2, er) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let ar = Splx.Val.cmp v1 v2 in
		let r =
			match ar, er with
			| 0, n when n = eq -> true
			| n1, n2 when n1 < 0 && n2 = lt -> true
			| n1, n2 when n1 > 0 && n2 = gt -> true
			| _, _ -> false
		in
		let to_s i = if i = 0 then "=" else if i < 0 then "<" else ">" in
		let e = (Splx.Val.pr v1) ^ " (expected: " ^ (to_s er) ^
			", actual: " ^ (to_s ar) ^ ")" in
		OUnit.assert_bool e r))
	in
	let tcs = [
		"z", Splx.Val.z, Splx.Val.z, eq;
		"zpd", Splx.Val.z, Splx.Val.pdelta Num.z, lt;
		"znd", Splx.Val.z, Splx.Val.ndelta Num.z, gt;
		"pdz", Splx.Val.pdelta Num.z, Splx.Val.z, gt;
		"ndz", Splx.Val.ndelta Num.z, Splx.Val.z, lt;
		"01", Splx.Val.of_num Num.z, Splx.Val.of_num Num.u, lt;
		"10", Splx.Val.of_num Num.u, Splx.Val.of_num Num.z, gt;
		"0-1", Splx.Val.of_num Num.z, Splx.Val.of_num Num.negU, gt;
		"-10", Splx.Val.of_num Num.negU, Splx.Val.of_num Num.z, lt;
		"12", Splx.Val.of_num Num.u, Splx.Val.of_num (Num.mk 2), lt;
		"21", Splx.Val.of_num (Num.mk 2), Splx.Val.of_num Num.u, gt;
		"-12", Splx.Val.of_num Num.negU, Splx.Val.of_num (Num.mk 2), lt;
		"2-1", Splx.Val.of_num (Num.mk 2), Splx.Val.of_num Num.negU, gt
	] in
	OUnit.TestList (List.map chk tcs)

(* Splx.Val.hasDelta *)
let hasDelta_ts =
	let chk (t, res, v) =
		OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_equal res (Splx.Val.hasDelta v)))
	in
	let tcs = [
		"z", false, Splx.Val.z;
		"int0", false, Splx.Val.of_num Num.u;
		"int1", false, Splx.Val.of_num (Num.mk 2);
		"int2", false, Splx.Val.of_num (Num.mk (-2));
		"rat0", false, Splx.Val.of_num (Num.mk ~den:3 2);
		"pdelta0", true, Splx.Val.pdelta Num.z;
		"pdelta1", true, Splx.Val.pdelta Num.u;
		"pdelta2", true, Splx.Val.adddelta (Splx.Val.pdelta (Num.mk 2));
		"pdelta3", true, Splx.Val.subdelta
			(Splx.Val.adddelta (Splx.Val.pdelta (Num.mk 2)));

		"ndelta0", true, Splx.Val.ndelta Num.z;
		"ndelta1", true, Splx.Val.subdelta (Splx.Val.ndelta Num.u);
		"no0", false, Splx.Val.adddelta (Splx.Val.ndelta Num.u);
		"no1", false, Splx.Val.subdelta (Splx.Val.pdelta Num.u)
	] in
	OUnit.TestList (List.map chk tcs)

(* Splx.Val.muln *)
let muln_ts =
	let chk (t, ar, er) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let estr = "ar = " ^ (Splx.Val.pr ar) ^ ", er = " ^ (Splx.Val.pr er) in
		OUnit.assert_bool estr (Splx.Val.cmp ar er = 0)))
	in
	let tcs = [
		"z", Splx.Val.muln Splx.Val.z Num.u, Splx.Val.z;
		"u", Splx.Val.muln (Splx.Val.of_num Num.u) Num.u,
			Splx.Val.of_num Num.u;
		"pu", Splx.Val.muln (Splx.Val.pdelta Num.u) Num.u,
			Splx.Val.pdelta Num.u;
		"nu", Splx.Val.muln (Splx.Val.ndelta Num.u) Num.u,
			Splx.Val.ndelta Num.u;
		"pn", Splx.Val.muln (Splx.Val.pdelta Num.u) (Num.mk 2),
			Splx.Val.add (Splx.Val.pdelta (Num.mk 2))
				(Splx.Val.pdelta Num.z);
	] in
	OUnit.TestList (List.map chk tcs)

(* Splx.Val.add *)
let add_ts =
	let chk (t, ar, er) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let estr = "ar = " ^ (Splx.Val.pr ar) ^ ", er = " ^ (Splx.Val.pr er) in
		OUnit.assert_bool estr (Splx.Val.cmp ar er = 0)))
	in
	let tcs = [
		"z", Splx.Val.add Splx.Val.z Splx.Val.z, Splx.Val.z;
		"u", Splx.Val.add Splx.Val.z (Splx.Val.of_num Num.u),
			Splx.Val.of_num Num.u;
		"pu", Splx.Val.add Splx.Val.z (Splx.Val.pdelta Num.u),
			Splx.Val.pdelta Num.u;
		"nu", Splx.Val.add Splx.Val.z (Splx.Val.ndelta Num.u),
			Splx.Val.ndelta Num.u;
		"pn", Splx.Val.add (Splx.Val.pdelta Num.u) (Splx.Val.ndelta Num.u),
			Splx.Val.of_num (Num.mk 2)
	] in
	OUnit.TestList (List.map chk tcs)

(* Splx.Val.adddelta *)
let adddelta_ts =
	let chk (t, i, o) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let r = Splx.Val.adddelta i in
		let e =
			"expected: " ^ (Splx.Val.pr o) ^
			", got: " ^ (Splx.Val.pr r)
		in
		OUnit.assert_bool e (Splx.Val.cmp o r = 0)))
	in
	let tcs = [
		"z", Splx.Val.z, Splx.Val.pdelta Num.z;
		"u", Splx.Val.of_num Num.u, Splx.Val.pdelta Num.u;
		"pos", Splx.Val.of_num (Num.mk 2), Splx.Val.pdelta (Num.mk 2);
		"ndelta", Splx.Val.ndelta Num.u, Splx.Val.of_num Num.u;
		"deltas", Splx.Val.pdelta Num.u,
			Splx.Val.add (Splx.Val.pdelta Num.u) (Splx.Val.pdelta Num.z)
	] in
	OUnit.TestList (List.map chk tcs)

(* Splx.Val.subdelta *)
let subdelta_ts =
	let chk (t, i, o) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let r = Splx.Val.subdelta i in
		let e =
			"expected: " ^ (Splx.Val.pr o) ^
			", got: " ^ (Splx.Val.pr r)
		in
		OUnit.assert_bool e (Splx.Val.cmp o r = 0)))
	in
	let tcs = [
		"z", Splx.Val.z, Splx.Val.ndelta Num.z;
		"u", Splx.Val.of_num Num.u, Splx.Val.ndelta Num.u;
		"pos", Splx.Val.of_num (Num.mk 2), Splx.Val.ndelta (Num.mk 2);
		"pdelta", Splx.Val.pdelta Num.u, Splx.Val.of_num Num.u;
		"deltas", Splx.Val.ndelta Num.u,
			Splx.Val.add (Splx.Val.ndelta Num.u) (Splx.Val.ndelta Num.z)
	] in
	OUnit.TestList (List.map chk tcs)

let ts =
	OUnit.TestList [
		OUnit.TestLabel ("z", z_ts);
		OUnit.TestLabel ("cmp", cmp_ts);
		OUnit.TestLabel ("hasDelta", hasDelta_ts);
		OUnit.TestLabel ("muln", muln_ts);
		OUnit.TestLabel ("add", add_ts);
		OUnit.TestLabel ("adddelta", adddelta_ts);
		OUnit.TestLabel ("subdelta", subdelta_ts)
	]
